"""Statistics module."""
from . import pam_stats
from . import running_stats
from . import significance

__all__ = []
__all__.extend(pam_stats.__all__)
__all__.extend(running_stats.__all__)
__all__.extend(significance.__all__)



"""Tests for the plots.scatter module."""
import numpy as np

from lmpy.matrix import Matrix
from lmpy.plots.scatter import create_scatter_plot


# .....................................................................................
def test_create_stat_scatter_plot(generate_temp_filename):
    """Test creating a scatter plot.

    Args:
        generate_temp_filename (pytest.Fixture): A fixture to generate filenames.
    """
    # Matrix to get scatter plot
    stat_matrix = Matrix(
        np.array(
            [
                [1, 2, 3],
                [1, 2, 3],
                [1, 2, 3],
                [1, 2, 3],
                [4, 5, 6],
                [7, 8, 9],
            ]
        ),
        headers={
            '0': [
                (0, 0.5, 0.5),
                (1, 0.5, 1.5),
                (2, 1.5, 0.5),
                (3, 1.5, 1.5),
                (4, 0.5, 0.5),
                (5, 0.5, 0.5),
            ],
            '1': ['stat_1', 'stat_2', 'stat_3']
        }
    )
    plot_filename = generate_temp_filename(suffix='.png')
    create_scatter_plot(
        plot_filename,
        stat_matrix[:, 1],
        stat_matrix[:, 2],
        title='Test Title',
        x_label='x label',
        y_label='y label',
        std_dev_styles=[
            dict(edgecolor='firebrick'),
            dict(edgecolor='fuchsia', linestyle='--'),
            dict(edgecolor='blue', linestyle=':'),
        ],
    )



"""Tool for wrangling a Matrix."""
import argparse

import json

from lmpy.data_wrangling.factory import WranglerFactory
from lmpy.species_list import SpeciesList
from lmpy.tools._config_parser import _process_arguments, get_logger


DESCRIPTION = 'Perform various wrangle operations on a species list.'


# .....................................................................................
def wrangle_species_list(species_list, wranglers):
    """Wrangle the SpeciesList as configured.

    Args:
        species_list (SpeciesList): A SpeciesList to wrangle.
        wranglers (list): A list of tree data wranglers.

    Returns:
        SpeciesList: The modified SpeciesList object.
        list: A list of report dictionaries.
    """
    report = []
    for wrangler in wranglers:
        species_list = wrangler.wrangle_species_list(species_list)
        report.append(wrangler.get_report())
    return species_list, report


# .....................................................................................
def build_parser():
    """Build an argparse.ArgumentParser object for the tool.

    Returns:
        argparse.ArgumentParser: An argument parser for the tool's parameters.
    """
    parser = argparse.ArgumentParser(prog='wrangle_matrix', description=DESCRIPTION)
    parser.add_argument('--config_file', type=str, help='Path to configuration file.')
    parser.add_argument(
        '--log_filename',
        '-l',
        type=str,
        help='A file location to write logging data.'
    )
    parser.add_argument(
        '--log_console',
        action='store_true',
        default=False,
        help='If provided, write log to console.'
    )
    parser.add_argument(
        '-r',
        '--report_filename',
        type=str,
        help='File location to write the wrangler report.'
    )
    parser.add_argument(
        'in_species_list_filename', type=str, help='Path to the input SpeciesList.'
    )
    parser.add_argument(
        'wrangler_configuration_file',
        type=str,
        help='Path to Matrix wrangler configuration.',
    )
    parser.add_argument(
        'out_species_list_filename', type=str, help='Path to the outut SpeciesList.'
    )
    return parser


# .....................................................................................
def cli():
    """Provide a command-line interface to the wrangle matrix tool."""
    parser = build_parser()
    args = _process_arguments(parser, config_arg='config_file')
    logger = get_logger(
        'wrangle_species_list',
        log_filename=args.log_filename,
        log_console=args.log_console
    )

    in_species_list = SpeciesList.from_file(args.in_species_list_filename)
    with open(args.wrangler_configuration_file, mode='rt') as in_json:
        wrangler_factory = WranglerFactory(logger=logger)
        wranglers = wrangler_factory.get_wranglers(json.load(in_json))

    wrangled_species_list, report = wrangle_species_list(in_species_list, wranglers)
    wrangled_species_list.write(args.out_species_list_filename)

    if args.report_filename is not None:
        with open(args.report_filename, mode='wt') as report_out:
            json.dump(report, report_out)


# .....................................................................................
__all__ = ['build_parser', 'cli', 'wrangle_species_list']


# .....................................................................................
if __name__ == '__main__':  # pragma: no cover
    cli()



"""Module containing tools for creating maps."""
import math

import matplotlib.image as mpimg
import matplotlib.pyplot as plt
import numpy as np

from lmpy.matrix import Matrix


# .....................................................................................
def create_map_matrix(min_x, min_y, max_x, max_y, resolution):
    """Creates an empty matrix to use for mapping.

    Args:
        min_x (numeric): The minimum x value of the heatmap range.
        min_y (numeric): The minimum y value of the heatmap range.
        max_x (numeric): The maximum x value of the heatmap range.
        max_y (numeric): The maximum y value of the heatmap range.
        resolution (numeric): The size of each matrix cell.

    Returns:
        Matrix: A Matrix of zeros for the spatial extent.
    """
    num_rows = math.ceil((max_y - min_y) / resolution)
    num_cols = math.ceil((max_x - min_x) / resolution)
    map_matrix = Matrix(
        np.zeros((num_rows, num_cols), dtype=int),
        headers={
            '0': [max_y - (j + .5) * resolution for j in range(num_rows)],
            '1': [min_x + (i + .5) * resolution for i in range(num_cols)]
        }
    )
    return map_matrix


# .....................................................................................
def get_row_col_for_x_y_func(min_x, min_y, max_x, max_y, resolution):
    """Get a function to return a row and column for an x, y.

    Args:
        min_x (numeric): The minimum x value of the heatmap range.
        min_y (numeric): The minimum y value of the heatmap range.
        max_x (numeric): The maximum x value of the heatmap range.
        max_y (numeric): The maximum y value of the heatmap range.
        resolution (numeric): The size of each matrix cell.

    Returns:
        Method: A method to generate row and column indices for an x, y.
    """
    num_rows = (max_y - min_y) // resolution
    num_cols = (max_x - min_x) // resolution

    # .......................
    def get_row_col_func(x, y):
        """Get the row and column where the point is located.

        Args:
            x (numeric): The x value to convert.
            y (numeric): The y value to convert.

        Returns:
            int, int: The row and column where the point is located.
        """
        r = int(min(num_rows - 1, max(0, int((max_y - y) // resolution))))
        c = int(min(num_cols - 1, max(0, int((x - min_x) // resolution))))
        return r, c

    return get_row_col_func


# .....................................................................................
def create_point_heatmap_matrix(readers, min_x, min_y, max_x, max_y, resolution):
    """Create a point heatmap matrix.

    Args:
        readers (PointReader or list of PointReader): A source of point data for
            creating the heatmap.
        min_x (numeric): The minimum x value of the heatmap range.
        min_y (numeric): The minimum y value of the heatmap range.
        max_x (numeric): The maximum x value of the heatmap range.
        max_y (numeric): The maximum y value of the heatmap range.
        resolution (numeric): The size of each matrix cell.

    Returns:
        Matrix: A matrix of point density.
    """
    heatmap = create_map_matrix(min_x, min_y, max_x, max_y, resolution)
    get_row_col_func = get_row_col_for_x_y_func(min_x, min_y, max_x, max_y, resolution)

    if not isinstance(readers, list):
        readers = [readers]
    for reader in readers:
        reader.open()
        for points in reader:
            for point in points:
                row, col = get_row_col_func(point.x, point.y)
                heatmap[row, col] += 1
        reader.close()
    return heatmap


# .....................................................................................
def create_stat_heatmap_matrix(matrix, stat, min_x, min_y, max_x, max_y, resolution):
    """Create a matrix stat heatmap matrix.

    Args:
        matrix (Matrix): A matrix containing site rows and stat columns.
        stat (str): The column header of the stat to create the map for.
        min_x (numeric): The minimum x value of the heatmap range.
        min_y (numeric): The minimum y value of the heatmap range.
        max_x (numeric): The maximum x value of the heatmap range.
        max_y (numeric): The maximum y value of the heatmap range.
        resolution (numeric): The size of each matrix cell.

    Returns:
        Matrix: A matrix of point density.
    """
    heatmap = create_map_matrix(min_x, min_y, max_x, max_y, resolution)
    num_hits = Matrix(np.zeros(heatmap.shape))
    get_row_col_func = get_row_col_for_x_y_func(min_x, min_y, max_x, max_y, resolution)

    matrix_column = matrix.get_column_headers().index(stat)
    site_headers = matrix.get_row_headers()
    for matrix_row in range(matrix.shape[0]):
        _, x, y = site_headers[matrix_row]
        row, col = get_row_col_func(x, y)
        heatmap[row, col] += matrix[matrix_row, matrix_column]
        num_hits[row, col] += 1

    # Get the mean value by dividing by the number of hits
    return np.divide(heatmap, num_hits, where=num_hits > 0)


# .....................................................................................
def plot_matrix(
    plot_filename,
    heatmap_matrix,
    base_layer=None,
    extent=None,
    mask_val=0,
    title=None,
    cmap=None,
    vmin=None,
    vmax=None,
):
    """Plot the heatmap.

    Args:
        plot_filename (str): A file path to write the figure image.
        heatmap_matrix (Matrix): A matrix of heatmap data (ordered like map grid).
        base_layer (str): A file location of a base layer map to add.
        extent (tuple): Image extent as (min_x, max_x, min_y, max_y) tuple.
        mask_val (numeric): A numeric value in the heatmap to ignore when mapping.
        title (str): A title to add to the image.
        cmap (str or colormap): The color map to use for the heat map.
        vmin (numeric): The minimum value to use when scaling color.
        vmax (numeric): The maximum value to use when scaling color.
    """
    if vmax == -1:
        vmax = int(heatmap_matrix.max())
        vmin = 0
    base_layer_params = {k: v for k, v in dict(extent=extent).items() if v is not None}
    heatmap_layer_params = {
        k: v for k, v in dict(
            extent=extent,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
        ).items() if v is not None
    }
    fig = plt.figure()
    ax = fig.add_subplot()
    if title is not None:
        ax.axes.set_title(title)
    ax.axes.xaxis.set_visible(False)
    ax.axes.yaxis.set_visible(False)

    if base_layer is not None:
        if isinstance(base_layer, str):
            ax.imshow(mpimg.imread(base_layer), **base_layer_params)

    ax.imshow(
        np.ma.masked_where(heatmap_matrix == mask_val, heatmap_matrix),
        **heatmap_layer_params
    )
    plt.savefig(plot_filename)


# .....................................................................................
__all__ = [
    'create_map_matrix',
    'create_point_heatmap_matrix',
    'create_stat_heatmap_matrix',
    'get_row_col_for_x_y_func',
    'plot_matrix',
]



"""Module containing occurrence data wranglers for matching a species list."""
from lmpy.data_wrangling.occurrence.base import _OccurrenceDataWrangler
from lmpy.species_list import SpeciesList


# .....................................................................................
class MatchSpeciesListWrangler(_OccurrenceDataWrangler):
    """Get an occurrence data wrangler for filtering by a species list."""
    name = 'OccurrenceSpeciesListWrangler'
    version = '1.0'

    # .......................
    def __init__(self, species_list, **params):
        """Get an occurrence data wrangler for filtering by bounding box.

        Args:
            species_list (SpeciesList): A species list to match against.
            **params (dict): Keyword parameters to pass to _OccurrenceDataWrangler.
        """
        _OccurrenceDataWrangler.__init__(self, **params)
        if isinstance(species_list, str):
            species_list = SpeciesList.from_file(species_list)
        self.species_list = species_list

    # .......................
    def _pass_condition(self, point):
        """Pass condition for a point.

        Args:
            point (Point): A point object to assess.

        Returns:
            bool: Indication if the point passed assessment.
        """
        return point.species_name in self.species_list



"""This tool encodes a phylogenetic tree into a matrix that can be used for MCPA."""
import argparse

from lmpy.data_preparation.tree_encoder import TreeEncoder
from lmpy.matrix import Matrix
from lmpy.tools._config_parser import _process_arguments
from lmpy.tree import TreeWrapper


DESCRIPTION = 'This tool encodes a phylogenetic tree into a matrix for use with MCPA.'


# .....................................................................................
def build_parser():
    """Build an argparse.ArgumentParser object for the tool.

    Returns:
        argparse.ArgumentParser: An argument parser for the tool's parameters.
    """
    parser = argparse.ArgumentParser(
        prog='my_prog_name',
        description=DESCRIPTION
    )
    # The '--config_file' argument is used for providing arguments in a configuration
    #     file instead of the command line
    parser.add_argument('--config_file', type=str, help='Path to configuration file.')
    parser.add_argument('tree_filename', type=str, help='Path to tree to encode.')
    parser.add_argument(
        'tree_schema',
        type=str,
        choices=['newick', 'nexus', 'nexml'],
        help='The schema of the provided tree filename.'
    )
    parser.add_argument(
        'pam_filename',
        type=str,
        help='File path to presence-absence matrix to use with the tree.'
    )
    parser.add_argument(
        'output_matrix_filename',
        type=str,
        help='File path where the encoded tree matrix should be written.'
    )
    return parser


# .....................................................................................
def cli():
    """Provide a command-line tool for encoding a tree for MCPA."""
    parser = build_parser()
    args = _process_arguments(parser, config_arg='config_file')
    tree = TreeWrapper.get(path=args.tree_filename, schema=args.tree_schema)
    pam = Matrix.load(args.pam_filename)
    tree_encoder = TreeEncoder(tree, pam)
    tree_mtx = tree_encoder.encode_phylogeny()
    tree_mtx.write(args.output_matrix_filename)


# .....................................................................................
__all__ = ['build_parser', 'cli']


# .....................................................................................
if __name__ == '__main__':  # pragma: no cover
    cli()



"""Module containing base class for data wranglers."""
import logging


# .....................................................................................
class _DataWrangler:
    """Data wrangler base class."""
    name = '_DataWrangler'
    version = '1.0'

    # ........................
    def __init__(self, *args, logger=None, **kwargs):
        """Base class constructor.

        Args:
            *args (tuple): Positional arguments.
            logger (logging.Logger): An optional logger to use for logging output.
            **kwargs (dict): Dictionary of parameters.
        """
        self.report = {}
        self.logger = logger

    # ........................
    @classmethod
    def from_config(cls, config):
        """Get an instance generated from the provided configuration.

        Args:
            config (dict): A dictionary of configuration parameters.

        Returns:
            _DataWrangler: The instance created from the parameters.
        """
        return cls(**config)

    # ........................
    def get_report(self):
        """Get a report of the results of the wrangler.

        Returns:
            dict: A dictionary of reporting information.
        """
        self.report['name'] = self.name
        self.report['version'] = self.version
        return self.report

    # ........................
    def log(self, msg, log_level=logging.INFO):
        """Log a message.

        Args:
            msg (str): A message to write to the logger.
            log_level (int): A level to use when logging the message.
        """
        if self.logger is not None:
            log_func = {
                logging.DEBUG: self.logger.debug,
                logging.INFO: self.logger.info,
                logging.WARNING: self.logger.warning,
                logging.ERROR: self.logger.error,
                logging.CRITICAL: self.logger.critical,
            }
            if log_level in log_func.keys():
                log_func[log_level](self.name + ': ' + msg)
            else:
                self.logger.log(log_level, self.name + ': ' + msg)



"""Calculate p-values for observed values compared to those generated from random."""
import argparse

from lmpy.matrix import Matrix
from lmpy.statistics.significance import (
    compare_absolute_values,
    compare_signed_values,
    get_significant_values,
    PermutationTests,
    SignificanceMethod,
)
from lmpy.tools._config_parser import _process_arguments


DESCRIPTION = 'Calculate p-values for observed data compared with random.'


# .....................................................................................
def build_parser():
    """Build an argparse.ArgumentParser object for the tool.

    Returns:
        argparse.ArgumentParser: An argument parser for the tool's parameters.
    """
    parser = argparse.ArgumentParser(
        prog='calculate_p_values',
        description=DESCRIPTION
    )
    # The '--config_file' argument is used for providing arguments in a configuration
    #     file instead of the command line
    parser.add_argument('--config_file', type=str, help='Path to configuration file.')
    parser.add_argument(
        '-m',
        '--significance_method',
        dest='significance_method',
        type=str,
        choices=['bonferroni', 'fdr', 'raw'],
        default='raw',
        help=(
            'If provided, correct the p-values using the Bonferroni method (bonferroni)'
            'or Benjamini and Hochberg method (fdr).'
        ),
    )
    parser.add_argument(
        '--num_permutations',
        '-n',
        type=int,
        help=(
            'The total number of random iterations used (useful if only computing'
            'p-values for a subset of randomizations).'
        ),
    )
    parser.add_argument(
        '--abs',
        action='store_true',
        help='Compare absolute values instead of signed values.'
    )
    parser.add_argument(
        '--alpha',
        type=float,
        default=0.05,
        help='Alpha value for statistical significance or False Discovery Rate.',
    )
    parser.add_argument(
        '-s',
        '--significance_matrix_filename',
        dest='significance_matrix_filename',
        type=str,
        help='Path to write significance matrix.'
    )
    parser.add_argument(
        'p_values_matrix',
        type=str,
        help='Path to write the computed p-values matrix.'
    )
    parser.add_argument(
        'observed_matrix',
        type=str,
        help='Path to the observed data matrix.',
    )
    parser.add_argument(
        'random_matrix',
        type=str,
        nargs='+',
        help='Path to the random data matrix.',
    )
    return parser


# .....................................................................................
def cli():
    """Provide a command-line tool for calculating p-values."""
    parser = build_parser()
    args = _process_arguments(parser, config_arg='config_file')
    # Compare method
    if args.abs:
        compare_func = compare_absolute_values
    else:
        compare_func = compare_signed_values
    # Load observed matrix
    obs = Matrix.load(args.observed_matrix)
    # Create running stats
    perm_testing = PermutationTests(obs, compare_fn=compare_func)
    # For each random matrix
    for rand_mtx_filename in args.random_matrix:
        # Load matrix
        rand_mtx = Matrix.load(rand_mtx_filename)
        # Add to running stats
        perm_testing.add_permutation(rand_mtx)
    # Get p-values, scaled if desired
    p_values = perm_testing.get_p_values(num_iterations=args.num_permutations)
    # If correction, do it
    # Write p-values matrix
    p_values.write(args.p_values_matrix)
    # Write significant matrix if desired
    if args.significance_matrix_filename is not None:
        if args.significance_method == 'raw':
            sig_method = SignificanceMethod.RAW
        elif args.significance_method == 'fdr':
            sig_method = SignificanceMethod.FDR
        else:
            sig_method = SignificanceMethod.BONFERRONI
        sig_mtx = get_significant_values(
            p_values,
            alpha=args.alpha,
            correction_method=sig_method,
        )
        sig_mtx.write(args.significance_matrix_filename)


# .....................................................................................
__all__ = ['build_parser', 'cli']


# .....................................................................................
if __name__ == '__main__':  # pragma: no cover
    cli()



"""This tool aggregates matrices by concatenating or adding."""
import argparse

from lmpy.matrix import Matrix
from lmpy.tools._config_parser import _process_arguments


DESCRIPTION = 'Aggregate matrices.'


# .....................................................................................
def build_parser():
    """Build an argparse.ArgumentParser object for the tool.

    Returns:
        argparse.ArgumentParser: An argument parser for the tool's parameters.
    """
    parser = argparse.ArgumentParser(
        prog='aggregate_matrices',
        description=DESCRIPTION
    )
    # The '--config_file' argument is used for providing arguments in a configuration
    #     file instead of the command line
    parser.add_argument('--config_file', type=str, help='Path to configuration file.')
    parser.add_argument(
        '--ndim',
        type=int,
        help='The number of dimensions in the output matrix for sum.',
        default=-1
    )
    parser.add_argument(
        'method',
        choices=['add', 'concatenate'],
        type=str,
        help='Method to use to aggregate the matrices.'
    )
    parser.add_argument(
        'axis',
        type=int,
        help='The axis to aggregate matrices on.'
    )
    parser.add_argument(
        'output_matrix_filename',
        type=str,
        help='Path to write out the aggregated matrices.'
    )
    parser.add_argument(
        'input_matrix_filename',
        type=str,
        nargs='+',
        help='Path to input matrix to aggregate.'
    )
    return parser


# .....................................................................................
def cli():
    """Provide a command-line tool for aggregating matrices."""
    parser = build_parser()
    args = _process_arguments(parser, config_arg='config_file')
    out_matrix = None
    input_matrices = [Matrix.load(fn) for fn in args.input_matrix_filename]
    if args.method == 'add':
        if args.ndim > 0:
            n_dims = args.ndim
        else:
            n_dims = min(mtx.ndim for mtx in input_matrices)
        for mtx in input_matrices:
            if mtx.ndim > n_dims:
                mtx = mtx.sum(axis=-1)
            if out_matrix is None:
                out_matrix = Matrix(mtx)
            else:
                out_matrix += mtx
    else:
        out_matrix = Matrix.concatenate(input_matrices, axis=args.axis)
    out_matrix.write(args.output_matrix_filename)


# .....................................................................................
__all__ = ['build_parser', 'cli']


# .....................................................................................
if __name__ == '__main__':  # pragma: no cover
    cli()



"""Module containing data wranglers for subsetting a species list to match a tree."""
from lmpy.tree import TreeWrapper

from lmpy.data_wrangling.species_list.base import _SpeciesListDataWrangler
from lmpy.species_list import SpeciesList


# .....................................................................................
class MatchTreeSpeciesListWrangler(_SpeciesListDataWrangler):
    """Subsets a species list to match the species in the tree."""
    name = 'MatchTreeSpeciesListWrangler'
    version = '1.0'

    # .......................
    def __init__(
        self,
        tree,
        **params
    ):
        """Constructor for MatchTreeSpeciesListWrangler class.

        Args:
            tree (TreeWrapper): A tree to match.
            **params (dict): Keyword parameters to pass to _SpeciesListDataWrangler.
        """
        _SpeciesListDataWrangler.__init__(self, **params)
        if isinstance(tree, str):
            tree = TreeWrapper.from_filename(tree)
        self.keep_names = [taxon.label for taxon in tree.taxon_namespace]

    # .......................
    def wrangle_species_list(self, species_list):
        """Wrangle a species list.

        Args:
            species_list (SpeciesList): A species list to wrangle.

        Returns:
            SpeciesList: A species list intersected with the provided tree.
        """
        ret_sl = SpeciesList(species_list.intersection(self.keep_names))
        self.report['removed'] = len(species_list) - len(ret_sl)
        self.log(f'removed {self.report["removed"]} tips.')
        return ret_sl



"""Create a point heatmap image."""
import argparse

from lmpy.plots.map import create_point_heatmap_matrix, plot_matrix
from lmpy.point import PointCsvReader, PointDwcaReader
from lmpy.tools._config_parser import _process_arguments


DESCRIPTION = 'Create a point map image.'


# .....................................................................................
def build_parser():
    """Build an argparse.ArgumentParser object for the tool.

    Returns:
        argparse.ArgumentParser: An argument parser for the tool's parameters.
    """
    parser = argparse.ArgumentParser(
        prog='create_point_heatmap',
        description=DESCRIPTION
    )
    # The '--config_file' argument is used for providing arguments in a configuration
    #     file instead of the command line
    parser.add_argument('--config_file', type=str, help='Path to configuration file.')
    parser.add_argument(
        '-b',
        '--base_layer',
        type=str,
        help='An optional base layer image for the plot.  Must have same extent.',
    )
    parser.add_argument(
        '--dwca',
        action='append',
        help='A Darwin-Core Archive filename to process.',
    )
    parser.add_argument(
        '--csv',
        action='append',
        nargs=4,
        help=(
            'A CSV file to process, a species header key, an x header key, '
            'and a y header key.'
        ),
    )
    parser.add_argument(
        '-t',
        '--title',
        type=str,
        help='A title to add to the plot image.'
    )
    parser.add_argument(
        '--cmap',
        type=str,
        default='Reds',
        help='A color map to use for the image.'
    )
    parser.add_argument(
        '--vmin',
        type=float,
        help='The minimum value for color scaling.'
    )
    parser.add_argument(
        '--vmax',
        type=float,
        help='The maximum value for color scaling (use -1 to use maximum from data).'
    )

    parser.add_argument(
        'plot_filename',
        type=str,
        help='A file path to write the generated plot image.',
    )
    parser.add_argument(
        'min_x',
        type=float,
        help='The minimum x value of the plot map range.'
    )
    parser.add_argument(
        'min_y',
        type=float,
        help='The minimum y value of the plot map range.'
    )
    parser.add_argument(
        'max_x',
        type=float,
        help='The maximum x value of the plot map range.'
    )
    parser.add_argument(
        'max_y',
        type=float,
        help='The maximum y value of the plot map range.'
    )
    parser.add_argument(
        'resolution',
        type=float,
        help='The resolution of each cell of the plot map.'
    )

    return parser


# .....................................................................................
def cli():
    """Provide a command-line tool for creating a point heatmap.

    Raises:
        ValueError: Raised if neither a CSV or DWCA is provided.
    """
    parser = build_parser()
    args = _process_arguments(parser, config_arg='config_file')
    readers = []
    if args.dwca is not None:
        for dwca_fn in args.dwca:
            readers.append(PointDwcaReader(dwca_fn))
    if args.csv is not None:
        for csv_fn, sp_key, x_key, y_key in args.csv:
            readers.append(PointCsvReader(csv_fn, sp_key, x_key, y_key))
    if len(readers) == 0:
        raise ValueError('Must provide at least one CSV and / or DWCA.')

    point_heatmap = create_point_heatmap_matrix(
        readers,
        args.min_x,
        args.min_y,
        args.max_x,
        args.max_y,
        args.resolution
    )
    plot_matrix(
        args.plot_filename,
        point_heatmap,
        base_layer=args.base_layer,
        extent=(args.min_x, args.max_x, args.min_y, args.max_y),
        mask_val=0,
        title=args.title,
        cmap=args.cmap,
        vmin=args.vmin,
        vmax=args.vmax,
    )


# .....................................................................................
__all__ = ['build_parser', 'cli']


# .....................................................................................
if __name__ == '__main__':  # pragma: no cover
    cli()



"""Module containing Species List Data Wrangler base class."""
from lmpy.data_wrangling.base import _DataWrangler


# .....................................................................................
class _SpeciesListDataWrangler(_DataWrangler):
    """Species List data wrangler base class."""
    name = '_SpeciesListDataWrangler'

    # .......................
    def __init__(self, **params):
        """Constructor for _SpeciesListDataWrangler base class.

        Args:
            **params (dict): Named parameters to pass to _DataWrangler base class.
        """
        _DataWrangler.__init__(self, **params)
        self.report['unresolved'] = 0
        self.report['duplicates'] = 0

    # .......................
    def wrangle_species_list(self, species_list):
        """Wrangle a species list.

        Args:
            species_list (SpeciesList): A species list to wrangle.

        Raises:
            NotImplementedError: This method is not implemeneted for the base class.
        """
        raise NotImplementedError('wrangle_species_list not implemented on base class.')



"""Module containing data wranglers for subsetting a tree to match a species list."""
from lmpy.species_list import SpeciesList

from lmpy.data_wrangling.tree.base import _TreeDataWrangler


# .....................................................................................
class MatchSpeciesListTreeWrangler(_TreeDataWrangler):
    """Subsets a tree to match the species in the species list."""
    name = 'MatchSpeciesListTreeWrangler'
    version = '1.0'

    # .......................
    def __init__(
        self,
        species_list,
        **params
    ):
        """Constructor for MatchSpeciesListTreeWrangler class.

        Args:
            species_list (SpeciesList): A species list to match.
            **params (dict): Keyword parameters to pass to _MatrixDataWrangler.
        """
        _TreeDataWrangler.__init__(self, **params)
        if isinstance(species_list, str):
            species_list = SpeciesList.from_file(species_list)
        self.keep_taxon_names = list(species_list)

    # .......................
    def wrangle_tree(self, tree):
        """Wrangle a tree.

        Args:
            tree (TreeWrapper): A tree to wrangle.

        Returns:
            TreeWrapper: The subsetted tree.
        """
        original_taxa_count = len(tree.taxon_namespace)
        tree.retain_taxa_with_labels(self.keep_taxon_names)
        tree.purge_taxon_namespace()
        num_purged = original_taxa_count - len(tree.taxon_namespace)
        self._purged += num_purged
        self.log(f'removed {self._purged} tips.')
        return tree



"""Module containing tools for evaluating statistical significance."""
from copy import deepcopy
from enum import Enum

import numpy as np

from lmpy import Matrix


# .....................................................................................
class SignificanceMethod(Enum):
    """Methods for determining statistical significance."""
    RAW = 0  # No significance value correction
    BONFERRONI = 1  # Bonferroni correction
    FDR = 2  # Benjamini and Hochberg correction


# .....................................................................................
def compare_absolute_values(observed, test_data):
    """Compares the absolute values of the observed and random data.

    Args:
        observed (:obj:`Numpy array`): A numpy array of observed values.
        test_data (:obj:`Numpy array`): A numpy array of random values.

    Returns:
        bool: Boolean indicating if the absolute value of the test data is greater than
            the absolute value of the observed data.
    """
    return np.abs(test_data) > np.abs(observed)


# .....................................................................................
def compare_signed_values(observed, test_data):
    """Compares the signed values of the observed and random data.

    Args:
        observed (:obj:`Numpy array`): A numpy array of observed values.
        test_data (:obj:`Numpy array`): A numpy array of random values.

    Returns:
        bool: An indication if the test data is greater than the observed data.
    """
    return test_data > observed


# .....................................................................................
def get_significant_values(
    p_values, alpha=0.05, correction_method=SignificanceMethod.RAW
):
    """Get significant values in a p-values matrix.

    Args:
        p_values (Matrix): A matrix of p-values to evaluate.
        alpha (float): An alpha value to use to evaluate significance.
        correction_method (int): The SignificanceMethod to use.

    Returns:
        Matrix: A boolean matrix indicating which values are cells are significant.

    Raises:
        ValueError: Raised if an unknown significance method is provided.
    """
    if correction_method == SignificanceMethod.RAW:
        return Matrix(
            p_values <= alpha,
            headers=p_values.get_headers(),
            metadata={'significance_method': 'Raw'},
        )
    elif correction_method == SignificanceMethod.BONFERRONI:
        return Matrix(
            np.minimum(p_values * p_values.size, 1.0),
            headers=p_values.get_headers(),
            metadata={'significance_method': 'Bonferroni'},
        )
    elif correction_method == SignificanceMethod.FDR:
        # In order to perform Benjamini and Hochberg correction
        # 1. Order p-values
        p_flat = p_values.flatten()
        num_vals = p_flat.size
        comp_p = 0.0
        # 2. Assign rank
        for rank, p_val in enumerate(sorted(list(p_flat))):
            # 3. Find the critical value where last p-val < (rank / num values) * alpha
            crit_val = alpha * ((rank + 1) / num_vals)
            if p_val < crit_val:
                comp_p = p_val
        # 4. All P(j) such that j <= i are significant
        return Matrix(
            p_values <= comp_p,
            headers=p_values.get_headers(),
            metadata={'significance_method': 'Benjamini and Hochberg'},
        )
    else:
        raise ValueError(f'Unknown significance method ({correction_method}).')


# .....................................................................................
class PermutationTests:
    """Tool for performing permutation tests."""
    # .......................
    def __init__(self, observed, compare_fn=compare_signed_values):
        """Constructor for PermutationTests class.

        Args:
            observed (Matrix): Observed values to test against.
            compare_fn (Method): A function used to compare values.
        """
        self.obs = observed
        self.f_counts = Matrix(np.zeros(self.obs.shape, dtype=int))
        self.count = 0
        self.compare_fn = compare_fn

    # .......................
    def add_permutation(self, test_val):
        """Compare the observed value with a test value and add to tracking.

        Args:
            test_val (Matrix): A matrix of permuted values to test against.
        """
        if test_val.ndim > self.obs.ndim:
            tmp_shp = list(self.obs.shape)
            tmp_shp.append(1)
            self.f_counts += np.sum(
                self.compare_fn(self.obs.reshape(tuple(tmp_shp)), test_val),
                axis=test_val.ndim - 1
            )
            self.count += test_val.shape[-1]
        else:
            self.f_counts += self.compare_fn(self.obs, test_val)
            self.count += 1

    # .......................
    def get_p_values(self, num_iterations=None):
        """Compute raw p-values from the permutations.

        Args:
            num_iterations (int): If provided, use this instead of self.count

        Returns:
            Matrix: A Matrix of p-values.
        """
        count = self.count
        if num_iterations is not None:
            count = num_iterations
        return Matrix(self.f_counts / count, headers=deepcopy(self.obs.get_headers()),)


# .....................................................................................
__all__ = [
    'compare_absolute_values',
    'compare_signed_values',
    'get_significant_values',
    'PermutationTests',
    'SignificanceMethod'
]



"""Module containing occurrence data wranglers for modifying a species list."""
from lmpy.data_wrangling.common.accepted_name_wrangler import (
    _AcceptedNameWrangler,
    resolve_names_gbif,
)
from lmpy.data_wrangling.species_list.base import _SpeciesListDataWrangler
from lmpy.species_list import SpeciesList


# .....................................................................................
class AcceptedNameSpeciesListWrangler(_SpeciesListDataWrangler, _AcceptedNameWrangler):
    """Modifies a species list to update taxon names to the "accepted" names."""
    name = 'AcceptedNameSpeciesListWrangler'
    version = '1.0'

    # .......................
    def __init__(
        self,
        name_map=None,
        name_resolver=None,
        purge_failures=True,
        out_map_filename=None,
        map_write_interval=100,
        out_map_format='json',
        **params
    ):
        """Constructor for AcceptedNameSpeciesListModifier class.

        Args:
            name_map (dict): A map of original name to accepted name.
            name_resolver (str or Method): If provided, use this method for getting new
                accepted names.  If set to 'gbif', use GBIF name resolution.
            purge_failures (bool): Should failures be purged from the tree.
            out_map_filename (str): A file location to write the updated name map.
            map_write_interval (int): Update the name map output file after each set of
                this many iterations.
            out_map_format (str): The format to write the names map (csv or json).
            **params (dict): Keyword parameters to pass to _TreeDataWrangler.
        """
        if isinstance(name_resolver, str) and name_resolver.lower() == 'gbif':
            name_resolver = resolve_names_gbif
        _AcceptedNameWrangler.__init__(
            self,
            name_map=name_map,
            name_resolver=name_resolver,
            out_map_filename=out_map_filename,
            map_write_interval=map_write_interval,
            out_map_format=out_map_format,
        )
        _SpeciesListDataWrangler.__init__(self, **params)

    # .......................
    def wrangle_species_list(self, species_list):
        """Wrangle a species list.

        Args:
            species_list (SpeciesList): A species list to wrangle.

        Returns:
            SpeciesList: A species list with accepted names.
        """
        accepted_species = set()
        acc_names = self.resolve_names(species_list)
        num_unresolved = 0
        num_duplicates = 0
        for name in species_list:
            if acc_names[name] is None:
                num_unresolved += 1
            elif acc_names[name] in accepted_species:
                num_duplicates += 1
            else:
                accepted_species.add(acc_names[name])
        self.report['unresolved'] = num_unresolved
        self.report['duplicates'] = num_duplicates
        return SpeciesList(accepted_species)



"""Create a scatter plot for two matrix vectors."""
import argparse
import json

from lmpy.matrix import Matrix
from lmpy.plots.scatter import create_scatter_plot
from lmpy.tools._config_parser import _process_arguments


DESCRIPTION = 'Create a scatter plot for two matrix statistics.'


# .....................................................................................
def build_parser():
    """Build an argparse.ArgumentParser object for the tool.

    Returns:
        argparse.ArgumentParser: An argument parser for the tool's parameters.
    """
    parser = argparse.ArgumentParser(
        prog='create_scatter_plot',
        description=DESCRIPTION
    )
    # The '--config_file' argument is used for providing arguments in a configuration
    #     file instead of the command line
    parser.add_argument('--config_file', type=str, help='Path to configuration file.')

    parser.add_argument('-t', '--title', type=str, help='A title for the plot image.')
    parser.add_argument('--x_label', type=str, help='A label for the x-axis.')
    parser.add_argument('--y_label', type=str, help='A label for the y-axis.')
    parser.add_argument(
        '-s',
        '--std_dev_style',
        type=str,
        action='append',
        help=(
            'A dictionary of parameters to style a standard deviation ellipse.  '
            'Multiple etries will result in multiple ellipses.'
        )
    )

    parser.add_argument(
        'plot_filename', type=str, help='File location to write plot image.'
    )
    parser.add_argument(
        'matrix_filename', type=str, help='File location of matrix containing data.'
    )
    parser.add_argument(
        'x_statistic',
        type=str,
        help='Column header of the vector to use for the x-axis.'
    )
    parser.add_argument(
        'y_statistic',
        type=str,
        help='Column header of the vector to use for the y-axis.'
    )
    return parser


# .....................................................................................
def cli():
    """Provide a command-line tool for creating a scatter plot."""
    parser = build_parser()
    args = _process_arguments(parser, config_arg='config_file')

    data = Matrix.load(args.matrix_filename)
    col_headers = data.get_column_headers()
    x_column = col_headers.index(args.x_statistic)
    y_column = col_headers.index(args.y_statistic)

    std_dev_styles = None
    if args.std_dev_style is not None:
        std_dev_styles = [json.loads(style) for style in args.std_dev_style]

    create_scatter_plot(
        args.plot_filename,
        data[:, x_column],
        data[:, y_column],
        title=args.title,
        x_label=args.x_label,
        y_label=args.y_label,
        std_dev_styles=std_dev_styles,
    )


# .....................................................................................
__all__ = ['build_parser', 'cli']


# .....................................................................................
if __name__ == '__main__':  # pragma: no cover
    cli()



"""Module containing methods for creating scatterplots."""
from matplotlib.patches import Ellipse
import matplotlib.pyplot as plt
import matplotlib.transforms as transforms
import numpy as np


# .....................................................................................
def confidence_ellipse(x, y, ax, n_std=3.0, facecolor='none', **kwargs):
    """Create a plot of the covariance confidence ellipse of *x* and *y*.

    From: https://matplotlib.org/stable/gallery/statistics/confidence_ellipse.html

    Args:
        x (array-like): Input data for the x dimension.
        y (array-likez): Input data for the y dimension.
        ax (matplotlib.axes.Axes): The axes object to draw the ellipse into.
        n_std (float): The number of standard deviations to determine the ellipse's
            radius.
        facecolor (str): A color argument for the Ellipse.
        **kwargs (dict): Keyword parameters forwarded to Ellipse.

    Returns:
        Ellipse: A matplotlib Ellipse.

    Raises:
        ValueError: Raised if sizes of x and y are not the same.
    """
    if x.size != y.size:
        raise ValueError("x and y must be the same size")

    cov = np.cov(x, y)
    pearson = cov[0, 1]/np.sqrt(cov[0, 0] * cov[1, 1])
    # Using a special case to obtain the eigenvalues of this
    # two-dimensional dataset.
    ell_radius_x = np.sqrt(1 + pearson)
    ell_radius_y = np.sqrt(1 - pearson)
    ellipse = Ellipse(
        (0, 0),
        width=ell_radius_x * 2,
        height=ell_radius_y * 2,
        facecolor=facecolor,
        **kwargs
    )

    # Calculating the standard deviation of x from
    # the square root of the variance and multiplying
    # with the given number of standard deviations.
    scale_x = np.sqrt(cov[0, 0]) * n_std
    mean_x = np.mean(x)

    # calculating the standard deviation of y ...
    scale_y = np.sqrt(cov[1, 1]) * n_std
    mean_y = np.mean(y)

    transf = transforms.Affine2D() \
        .rotate_deg(45) \
        .scale(scale_x, scale_y) \
        .translate(mean_x, mean_y)

    ellipse.set_transform(transf + ax.transData)
    return ax.add_patch(ellipse)


# .....................................................................................
def create_scatter_plot(
    plot_filename,
    x_data,
    y_data,
    title=None,
    x_label=None,
    y_label=None,
    std_dev_styles=None,
    legend=True
):
    """Create a scatter plot for two data vectors with optional std dev ellipses.

    Args:
        plot_filename (str): A file path to write the plot image.
        x_data (array-like): Input data for the x dimension.
        y_data (array-like): Input data for the y dimension.
        title (str): A title to add to the plot.
        x_label (str): A label for the x-axis of the plot.
        y_label (str): A label for the y-axis of the plot.
        std_dev_styles (list of dict): A list of dictionaries of keyword arguments to
            use for stylize each confidence ellipse.  A confidence of ellipse will be
            added for each dictionary.
        legend (bool): Should a legend be added to the image.
    """
    scatter_args = dict(c='#0000ff', zorder=1, s=10)
    fig, axs = plt.subplots(1, 1)

    if std_dev_styles is not None:
        for i, style_dict in enumerate(std_dev_styles):
            confidence_ellipse(
                x_data, y_data, axs, n_std=i+1, label=rf'${i}\sigma$', **style_dict
            )

    axs.scatter(x_data, y_data, **scatter_args)

    if title is not None:
        axs.set_title(title)
    if x_label is not None:
        axs.set_xlabel(x_label)
    if y_label is not None:
        axs.set_ylabel(y_label)

    if legend:
        axs.legend()

    plt.savefig(plot_filename)
    fig.clear()


# .....................................................................................
__all__ = ['confidence_ellipse', 'create_scatter_plot']



"""Tests for the plots.map module."""
import numpy as np

from lmpy.matrix import Matrix
from lmpy.plots.map import (
    create_point_heatmap_matrix,
    create_stat_heatmap_matrix,
    plot_matrix,
)
from lmpy.point import Point, PointCsvReader, PointCsvWriter


# .....................................................................................
def test_create_point_heatmap_matrix_known_data(generate_temp_filename):
    """Test creating a matrix heatmap from a point reader.

    Args:
        generate_temp_filename (pytest.Fixture): A fixture to generate filenames.
    """
    # Generate and write points
    point_filename = generate_temp_filename(suffix='.csv')
    points = [
        Point('Species A', 1, 3),
        Point('Species A', 2, 3),
        Point('Species A', 1.25, 3.5),
        Point('Species A', 3, 2),
        Point('Species A', 4, 1),
        Point('Species A', 5, 0.3),
        Point('Species A', 5.9, 5.4),
        Point('Species A', 0.1, 4.5),
        Point('Species A', 1.1, 5),
        Point('Species A', 2.3, 3.5),
        Point('Species A', 1, 3),
    ]
    with PointCsvWriter(point_filename, ['species_name', 'x', 'y']) as writer:
        writer.write_points(points)

    # Open point reader
    reader = PointCsvReader(point_filename, 'species_name', 'x', 'y')
    reader.open()

    # Create heat map
    heatmap = create_point_heatmap_matrix(reader, 0, 0, 6, 6, 2)
    reader.close()

    # Test that the heat map looks like what we expect
    test_heatmap = Matrix(
        np.array(
            [
                [2, 0, 1],
                [3, 2, 0],
                [0, 1, 2]
            ]
        ),
        headers={
            '0': [1, 3, 5],
            '1': [1, 3, 5]
        }
    )
    assert np.all(test_heatmap == heatmap)


# .....................................................................................
def test_create_point_heatmap_matrix_random_data(generate_temp_filename):
    """Test creating a matrix heatmap from a point reader.

    Args:
        generate_temp_filename (pytest.Fixture): A fixture to generate filenames.
    """
    # Generate and write points
    point_filename = generate_temp_filename(suffix='.csv')
    num_points = np.random.randint(20, 1000)
    with PointCsvWriter(point_filename, ['species_name', 'x', 'y']) as writer:
        for _ in range(num_points):
            writer.write_points(
                [
                    Point(
                        'Species A',
                        np.random.randint(1, 99),
                        np.random.randint(1, 99)
                    )
                ]
            )

    # Open point reader
    reader = PointCsvReader(point_filename, 'species_name', 'x', 'y')
    reader.open()

    # Create heat map
    heatmap = create_point_heatmap_matrix(reader, 0, 0, 100, 100, 2)
    reader.close()

    assert heatmap.sum() == num_points


# .....................................................................................
def test_create_stat_heatmap_matrix():
    """Test creating a statistic heatmap."""
    # Matrix to get heatmap for
    stat_matrix = Matrix(
        np.array(
            [
                [1, 2, 3],
                [1, 2, 3],
                [1, 2, 3],
                [1, 2, 3],
                [4, 5, 6],
                [7, 8, 9],
            ]
        ),
        headers={
            '0': [
                (0, 0.5, 0.5),
                (1, 0.5, 1.5),
                (2, 1.5, 0.5),
                (3, 1.5, 1.5),
                (4, 0.5, 0.5),
                (5, 0.5, 0.5),
            ],
            '1': ['stat_1', 'stat_2', 'stat_3']
        }
    )
    stat_heatmap = create_stat_heatmap_matrix(stat_matrix, 'stat_2', 0, 0, 2, 2, 1)
    test_heatmap_matrix = Matrix(
        np.array(
            [
                [2, 2],
                [5, 2]
            ],
        ),
        headers={
            '0': [0.5, 1.5],
            '1': [0.5, 1.5]
        }
    )

    assert np.all(stat_heatmap == test_heatmap_matrix)


# .....................................................................................
def test_plot_matrix_points(generate_temp_filename):
    """Test the plot_matrix function with points.

    Args:
        generate_temp_filename (pytest.Fixture): A fixture to generate filenames.
    """
    # Generate and write points
    point_filename = generate_temp_filename(suffix='.csv')
    num_points = np.random.randint(20, 1000)
    with PointCsvWriter(point_filename, ['species_name', 'x', 'y']) as writer:
        for _ in range(num_points):
            writer.write_points(
                [
                    Point(
                        'Species A',
                        np.random.randint(1, 99),
                        np.random.randint(1, 99)
                    )
                ]
            )

    # Open point reader
    reader = PointCsvReader(point_filename, 'species_name', 'x', 'y')
    reader.open()

    # Create heat map
    heatmap = create_point_heatmap_matrix(reader, 0, 0, 100, 100, 2)
    reader.close()

    plot_filename = generate_temp_filename(suffix='.png')
    plot_matrix(plot_filename, heatmap)



"""Module containing data wranglers for matching a matrix to a species list."""
from lmpy.species_list import SpeciesList

from lmpy.data_wrangling.matrix.base import _MatrixDataWrangler


# .....................................................................................
class MatchSpeciesListMatrixWrangler(_MatrixDataWrangler):
    """Subset and reorder a matrix to match a species list."""
    name = 'MatchSpeciesListMatrixWrangler'
    version = '1.0'

    # .......................
    def __init__(
        self,
        species_list,
        species_axis,
        **params
    ):
        """Constructor for MatchSpeciesListMatrixWrangler class.

        Args:
            species_list (SpeciesList or str): A species list object or file path.
            species_axis (int): The matrix axis with species names to match.
            **params (dict): Keyword parameters to pass to _MatrixDataWrangler.
        """
        _MatrixDataWrangler.__init__(self, **params)

        if isinstance(species_list, str):
            species_list = SpeciesList.from_file(species_list)
        self.species_list = species_list
        self.species_axis = species_axis
        self.report['changes'] = {}

    # .......................
    def wrangle_matrix(self, matrix):
        """Wrangle a matrix.

        Args:
            matrix (Matrix): A matrix to wrangle.

        Returns:
            Matrix: The matrix subset to species list entries.
        """
        slices = []
        for axis in range(matrix.ndim):
            if axis == self.species_axis:
                axis_slice = []
                axis_headers = matrix.get_headers(axis=str(axis))

                for name in self.species_list:
                    if name in axis_headers:
                        axis_slice.append(axis_headers.index(name))

                if str(axis) not in self.report['changes'].keys():
                    self.report['changes'][str(axis)] = {'purged': 0}
                self.report[
                    'changes'
                ][str(axis)]['purged'] += (len(axis_headers) - len(axis_slice))
                self.log('Purged {} species.'.format(
                    self.report['changes'][str(axis)]['purged'])
                )
            else:
                axis_slice = list(range(matrix.shape[axis]))

            slices.append(axis_slice)

        return matrix.slice(*slices)



"""A tool for wrangling phylogenetic trees."""
import argparse
from copy import deepcopy
import json

from lmpy.data_wrangling.factory import WranglerFactory
from lmpy.tools._config_parser import _process_arguments, get_logger
from lmpy.tree import TreeWrapper


# .....................................................................................
DESCRIPTION = '''A tool to wrangle phylogenetic trees as needed for processing.'''


# .....................................................................................
def wrangle_tree(tree, wranglers):
    """Wrangle the tree as configured.

    Args:
        tree (TreeWrapper): A phylogenetic tree to wrangle.
        wranglers (list): A list of tree data wranglers.

    Returns:
        TreeWrapper: The modified tree object.
        dict: Wrangler reporting information
    """
    report = []
    for wrangler in wranglers:
        tree = wrangler.wrangle_tree(tree)
        report.append(deepcopy(wrangler.get_report()))
    return tree, report


# .....................................................................................
def build_parser():
    """Build an argparse.ArgumentParser object for the tool.

    Returns:
        argparse.ArgumentParser: An argument parser for the tool's parameters.
    """
    parser = argparse.ArgumentParser(prog='wrangle_tree', description=DESCRIPTION)
    parser.add_argument('--config_file', type=str, help='Path to configuration file.')
    parser.add_argument(
        '--log_filename',
        '-l',
        type=str,
        help='A file location to write logging data.'
    )
    parser.add_argument(
        '--log_console',
        action='store_true',
        default=False,
        help='If provided, write log to console.'
    )
    parser.add_argument(
        '-r', '--report_filename', type=str, help='Path to write report.'
    )
    parser.add_argument('tree_filename', type=str, help='Path to phylogenetic tree.')
    parser.add_argument(
        'tree_schema',
        type=str,
        choices=['newick', 'nexus'],
        help='The schema of the input phylogenetic tree.',
    )
    parser.add_argument(
        'wrangler_configuration_file',
        type=str,
        help='Path to phylogenetic tree wrangler configuration.',
    )
    parser.add_argument(
        'out_tree_filename',
        type=str,
        help='Path to write the wrangled phylogenetic tree.',
    )
    parser.add_argument(
        'out_tree_schema',
        type=str,
        choices=['newick', 'nexus'],
        help='The schema of the output phylogenetic tree.',
    )
    return parser


# .....................................................................................
def cli():
    """Command-line interface for the tool."""
    parser = build_parser()
    args = _process_arguments(parser, config_arg='config_file')
    logger = get_logger(
        'wrangle_tree',
        log_filename=args.log_filename,
        log_console=args.log_console
    )
    tree = TreeWrapper.get(path=args.tree_filename, schema=args.tree_schema)
    with open(args.wrangler_configuration_file, mode='rt') as in_json:
        wrangler_factory = WranglerFactory(logger=logger)
        wranglers = wrangler_factory.get_wranglers(json.load(in_json))
    wrangled_tree, report = wrangle_tree(tree, wranglers)
    wrangled_tree.write(path=args.out_tree_filename, schema=args.out_tree_schema)
    if args.report_filename:
        with open(args.report_filename, mode='wt') as report_out:
            json.dump(report, report_out)


# .....................................................................................
__all__ = ['cli', 'wrangle_tree']


# .....................................................................................
if __name__ == '__main__':  # pragma: no cover
    cli()



"""Module containing species data wranglers for creating a union species list."""
from lmpy.data_wrangling.species_list.base import _SpeciesListDataWrangler
from lmpy.species_list import SpeciesList


# .....................................................................................
class UnionSpeciesListWrangler(_SpeciesListDataWrangler):
    """Modifies a species list by creating a union with another."""
    name = 'UnionSpeciesListWrangler'
    version = '1.0'

    # .......................
    def __init__(
        self,
        species_list,
        **params
    ):
        """Constructor for UnionSpeciesListModifier class.

        Args:
            species_list (SpeciesList): The species list to union with.
            **params (dict): Keyword parameters to pass to _TreeDataWrangler.
        """
        if isinstance(species_list, str):
            species_list = SpeciesList.from_file(species_list)
        self.other_species_list = species_list
        _SpeciesListDataWrangler.__init__(self, **params)

    # .......................
    def wrangle_species_list(self, species_list):
        """Wrangle a species list.

        Args:
            species_list (SpeciesList): A species list to wrangle.

        Returns:
            SpeciesList: A species list union with the provided species list.
        """
        ret_sl = SpeciesList(species_list.union(self.other_species_list))
        self.report['added'] = len(ret_sl) - len(species_list)
        self.log(f'added {self.report["added"]} tips.')
        return ret_sl



"""Module containing SpeciesList class."""


# .....................................................................................
class SpeciesList(set):
    """A class containing a list of species for analysis."""
    # .......................
    def __init__(self, *args, **kwargs):
        """Constructor for SpeciesList.

        Args:
            *args (list of object): A list of positional arguments to send to `list`.
            **kwargs (dict): A dictionary of keyword arguments to send to `list`.
        """
        super().__init__(*args, **kwargs)

    # .......................
    @classmethod
    def from_file(cls, filename):
        """Get a species list from a file.

        Args:
            filename (str): File path of a species list to load.

        Returns:
            SpeciesList: A species list loaded from a file.
        """
        with open(filename, mode='rt') as in_file:
            species = [line.replace('\n', '').strip() for line in in_file]
        return SpeciesList(species)

    # .......................
    def write(self, filename):
        """Write the species list to a file.

        Args:
            filename (str): Path where the species list data should be written.
        """
        with open(filename, mode='wt') as out_file:
            for name in self:
                out_file.write(f'{name}\n')



"""Species List data wranglers."""

from . import accepted_name_wrangler
from . import base
from . import intersect_species_list_wrangler
from . import match_matrix_wrangler
from . import match_tree_wrangler
from . import union_species_list_wrangler

__all__ = [
    'accepted_name_wrangler',
    'base',
    'intersect_species_list_wrangler',
    'match_matrix_wrangler',
    'match_tree_wrangler',
    'union_species_list_wrangler',
]
